(**TD06**)
open T
open Type

exception Error of string

type eqs = (Type.t * Type.t) list
type subst = (int * Type.t) list

(***********************)
(* Quelques afficheurs *)
(***********************)

open  Printf

let print_eq chan (t1,t2) =
  fprintf chan "%a = %a" Type.print t1 Type.print t2

let print_eqs chan eqs =
  List.iter
    (fun eq -> fprintf chan "%a\n" print_eq eq) 
    eqs ;
  flush chan

let print_subst chan s =
  print_eqs chan
    (List.map (fun (n,t) -> Tvar n,t) s) ;
  flush chan


(***************************************)
(* Algorithme d'unification de Robison *)
(***************************************)

(* occur_check, comme le nom l'indique *)
let rec occur_check n t = match t with
| Nat -> ()
| Arrow (t1,t2) -> occur_check n t1 ; occur_check n t2
| Tvar m -> if n=m then raise (Error "Occur")

(** Appliquer une substitution  une variable *)
let subst s x = try List.assoc x s with Not_found -> Tvar x

(** 'subst_type s t' applique la substitution 's' au type 't' *)
let rec subst_type s t = match t with
  | Nat -> Nat
  | Tvar x -> subst s x
  | Arrow (t1, t2) ->
      Arrow (subst_type s t1, subst_type s t2)

(** Appliquer une substitution  un ensemble d'quations *)
let rec subst_eqs s eqs =
  List.map (fun (t1,t2) -> subst_type s t1, subst_type s t2) eqs

(** Et  une substitution *)
let rec subst_subst s0 s1 =
  List.map (fun (x,t) -> x, subst_type s0 t) s1


(*****************************************)
(* Algorithme de Robinson proprement dit *)
(*****************************************)

(*
   Les arguments de unif :
  - eqs est la liste des quations non-examines.
  - r est la liste des quations examines (sous forme de substitution)
*)
let rec unif eqs r = match eqs with
| [] -> r
| (t1,t2)::eqs -> unif_rec t1 t2 eqs r

(* Traiter l'quation t1=t2, il reste eqs  examiner,
   r sont dj examines *)
and unif_rec t1 t2 eqs r = match t1,t2 with
(* Chouette or-pat *)
| (Nat,Arrow (_,_))|(Arrow (_,_),Nat) -> raise (Error "Nat against A->B")
| Arrow (t1,t2),Arrow (t3,t4) ->
    unif_rec t1 t3 ((t2,t4)::eqs) r
| Nat,Nat -> unif eqs r
(* Noter la garde 'when' *)
| Tvar n, Tvar m when n=m -> unif eqs r
(* Encore un or-pat, encore plus chouette *)
| (Tvar n,t)|(t,Tvar n) ->
    occur_check n t ;
    (* Tout est l : bien prendre garde de remplacer n par t PARTOUT *)
    unif (subst_eqs [n,t] eqs) ((n,t)::subst_subst [n,t] r)

(* 'verbose' est 'true' si on a donn l'option "-v" sur la ligne de commande *)
let verbose = List.mem "-v" (Array.to_list Sys.argv)

let mgu eqs =
  if verbose then begin
    prerr_endline "*** EQS ***" ;
    print_eqs stderr eqs
  end ;
  let s = unif eqs [] in
  if verbose then begin
    prerr_endline "*** SUBST ***" ;
    print_subst stderr s ; flush stderr
  end ;
  s

(* Un truc simple pour avoir des variables  fraches  *)
let fresh_var =
  let c = ref 0 in
  fun () -> let r = !c in incr c ; r

let mgu_incr t1 t2 s =
  if verbose then begin
    prerr_endline "*** EQ ***" ;
    print_eq stderr (t1,t2) ;
    eprintf "\n + SUBST\n" ;
    print_subst stderr s ;
    flush stderr
  end ;
  let s = unif_rec (subst_type s t1) (subst_type s t2) [] s in
  if verbose then begin
    prerr_endline "*** RESULT ***" ;
    print_subst stderr s ;
    prerr_endline "** END **"
  end ;
  s

(**END**)

(** TD 7 **)


module Vars = struct
  module V = Set.Make
  (struct
    type t = int
    let compare i1 i2 =
      if i1 < i2 then -1
      else if i1 > i2 then 1
      else 0
  end)
  include V

  let of_list s = List.fold_right add s empty
  let disjoint s1 s2 = is_empty (inter s1 s2)
end


(* Domaine d'une substitution *)
let dom s = Vars.of_list (List.map fst s)

(* Quelques substitutions simples *)
let id = []
and delta x t = [x,t]

(* Et composition des substitutions:   
   (s0 o s1)(x) = s0(s1(x)) si $x \in dom(s1)$
   (s0 o s1)(x) = s0(x)      si $x \not\in dom(s1)$. *)

let comp s0 s1 =
(* On suppose (et vrifie) que les supports sont disjoints *)
  assert (Vars.disjoint  (dom s0) (dom s1)) ;
  subst_subst s0 s1 @ s0

type scheme = int list * Type.t

(* Affichage des schma de types avec des noms
   de variables de la forme "A", "B",... *)

let print_vars chan =
  let rec p_rec chan xs = match xs with
  | [] -> ()
  | [_,x] -> fprintf chan "%s" x
  | (_,x)::xs -> fprintf chan "%s,%a" x p_rec xs in
  fun xs -> fprintf chan "(%a)" p_rec xs

let rec p_type xs chan t = match t with
| Nat -> fprintf chan "Nat"
| Arrow (t1,t2) ->
    fprintf chan "%a -> %a" (p_type_arg xs) t1 (p_type xs) t2
| Tvar x ->
    begin try
      fprintf chan "%s" (List.assoc x xs)
    with Not_found -> fprintf chan "t%i" x
    end

and p_type_arg xs chan t = match t with
| Arrow _ -> fprintf chan "(%a)" (p_type xs) t
| _ -> p_type xs chan t


exception No

let next_char c = match c with
| 'Z' -> raise No
| _   -> Char.chr (Char.code c+1)

let next v =
  let rec do_rec len =
    if len = 0 then "A"
    else try
      let c = next_char v.[len-1] in
      String.sub v 0 (len-1) ^ String.make 1 c
    with No -> do_rec (len-1) ^ "A" in
  do_rec (String.length v)


let rec normalize xs t = match t with
| Nat -> []
| Tvar x -> if List.mem x xs then [x] else []
| Arrow (t1,t2) ->
    let xs1 = normalize xs t1 and xs2 = normalize xs t2 in
    xs1 @ List.filter (fun x -> not (List.mem x xs1)) xs2

let print_scheme chan (xs,ty) = match xs with
| [] -> Type.print chan ty
| _  ->
    let rec do_rec v xs = match xs with
    | [] -> []
    | x::xs ->
	let v = next v in
	(x,v)::do_rec v xs in
    let xs = do_rec "" (normalize xs ty) in
    fprintf chan "%a[%a]" print_vars xs (p_type xs) ty


