(** PCF objet *)

module Ast : sig
(** Arbres de syntaxe abstraite pour PCF objet *)

(** Les tiquettes des mthodes *)
type label = string

(** Type de la syntaxe abstraite des termes de oPCF. *)
type t =
  | Num of int               (** Constante entire *)
  | Var of string            (** Accs  une variable *)
  | Op  of Op.t * t * t        (** Application d'un oprateur [t1 op t2] *)
  | Ifz of t * t * t         (** Conditionnelle [Ifz t1 Then t2 Else t3] *)
  | Let of string * t * t    (** Liaison locale [Let x = t1 In t2] *) 
  | App of t * t             (** Application de fonction [t1 t2] *)
  | Fun of string * t        (** Dfinition de fonction [Fun x -> t] *)
  | Fix of string * t        (** Point fixe [Fix x -> t] *)
  | Rec      of (label * t) list      (** Cration d'un enregistrement [{ ...; lab=t; ..}] *)
  | Access   of t * label             (** Accs  un champ d'enregistrement [t.lab]*)
  | CopyWith of t * (label * t) list  (** Copie/extension d'enregistrement [{ t with  ... ; lab=t; ... }]  *)
(** Objets *)
  | Obj of string option *  (label * t) list (** Cration d'un objet [obj [this] { ... ; lab=t ; ... }] *)
  | Invoke of t * label (** Invocation de mthode [t#lab] *)
  | ObjWith of string option  * t * (label * t) list (** Extension d'un objet [obj [this] { t with ...; lab=t ; }] *)

end

(** Lecture des termes PCF + objets,  comme {!S.Reader}. *)

module Reader : sig

(** Dans un canal *)  
  val from_chan : in_channel -> Ast.t

(** Dans une chane *)
  val from_string : string -> Ast.t
end

(** Boucle interactive  pour PCF + objets, comme {!S.Top}. *)
module Top : sig

val loop :  (out_channel -> 'a -> unit) -> ('a Env.t -> Ast.t -> 'a) -> unit
(** [loop print eval] lance une boucle 'toplevel', en franais boucle interactive.
    Une telle boucle permet d'valuer une srie de termes PCF dans
    un environnement qui va croissant.

    Une fois lance la boucle (par [loop print eval]), on voit une invite
    du style {b PCF> }. En rponse  cette invite on peut crire,
      - Un terme PCF (suivi d'un double point virgule) [t;;].
      - Ou une dfinition ordinaire [Let x = t;;].
      - Ou une dfinition rcursive [Let Rec x = t;;],
        (qui est lue comme [Let x = Fix x -> t;;], ce qui nous ramne au
         cas prcdent).

    Dans tous les cas, la boucle appelle la fonction [eval] fournie
    comme [eval env t], o
      - le terme PCF [t] est donn ci-dessus,
      - [env] est un environnement initialement vide.

    Cet appel renvoie un rsulat [r] de type ['a], qui est
    affich par la fonction [print] fournie.

    Enfin, la boucle interactive se rappelle, mais avec un
    environnement tendu par la liaison de [x]  [r] dans le
    cas o l'on avait crit [Let x = t;;].

   L'effet gnral est celui de la boucle interactive de Caml.
   @see <../index.html#top> Exemple d'utilisation de [Toop.loop].
*)

end


